// TFTP Sliding Window Client
// Jacob M. Peck
// CSC 445 - Lea

// note: some code adapted from: 
// http://thushw.blogspot.com/2011/06/asynchronous-udp-server-using-java-nio.html

import java.nio.*;
import java.nio.channels.*;
import java.util.*;
import java.io.*;
import java.net.*;

public class Client extends TFTPConstants {
  public int mode = MODE_WAIT;
  public List<byte[]> data = new ArrayList<byte[]>();
  public boolean[] sentframe;
  public String filename = "";
  public InetSocketAddress server_addr;
  public String hostname;
  public boolean ipv6 = false;
  public boolean fakedrop = false;
  public final double DROPRATE = 0.01;
  public Random random = new Random();
  public int len = 0;
  public int totallen = 0;
  public long starttime = 0;
  public long endtime = 0;
  public long totaltime = 0;
  public Selector selector;
  public SelectionKey selection_key;
  
  public static void main(String[] args) {
    Client c = new Client();
    JEclap jeclap = new JEclap(args);
    // args
    if(jeclap.hasArg("6")) c.ipv6 = true; // ipv6
    if(jeclap.hasArg("d")) c.fakedrop = true; // fake drop
    if(jeclap.hasArg("s")) c.SWS = 8; // sliding window
    c.hostname = jeclap.getArg("h");
    //c.server_addr = new InetSocketAddress(jeclap.getArg("h"), INITIAL_TID);
    c.run();
  }
  
  public void loadFile(File f) throws Exception {
    data = new ArrayList<byte[]>();
    FileInputStream in = new FileInputStream(f);
    filename = f.getName();
    len = (int)(f.length());
    //System.out.println("len = " + len);
    totallen += len;
    for(int i = 0; i < len/BLOCK_SIZE + 1; i++) {
      byte[] filedata;
      if(i == len/BLOCK_SIZE)
        filedata = new byte[len - (BLOCK_SIZE * i-1) - 1];
      else
       filedata = new byte[BLOCK_SIZE];
      for(int j = 0; j < filedata.length; j++) {
        try {
          filedata[j] = (byte)in.read();
        } catch (Exception ex) {
          break;
        }
      }
      data.add(filedata);
    }
    
    //System.out.println(data.size());
    
    sentframe = new boolean[data.size()];
    for(int i = 0; i < sentframe.length; i++) sentframe[i] = false;
  }
  
  public void send() throws Exception {
    initializeSender();
    starttime = System.nanoTime();
    while(mode != MODE_DONE) {
      //System.out.println("Waiting on events.  Mode " + mode);
      selector.select(TIMEOUT);
      for(SelectionKey sk : selector.selectedKeys()) {
        selector.selectedKeys().remove(sk);
        
        if(!sk.isValid()) continue; // bad key
        
        handleKey(sk); // handle the key
      }
      toggleInterestOps();
      selection_key.interestOps(INTEREST_OPS);
    }
    endtime = System.nanoTime();
    //System.out.println("Exiting...");
    
    totaltime += (endtime - starttime);
    
    float millis = (endtime - starttime) / 1000000;
    int bits = len * 8;
    float numerator = bits * 1000;
    float denominator = millis * 1000000;
    System.out.println("Time taken (ms): " + millis);
    System.out.println("Throughput (Mbps): " + (numerator / denominator)); 
  }
    
  
  public void run() {
    try {
      System.out.println("Initializing...");
      initializeSender();
      maintainSenderInvariant();
      
      selector = Selector.open();
      DatagramChannel data_channel = null;
      data_channel = ipv6 ? DatagramChannel.open(StandardProtocolFamily.INET6)
                          : DatagramChannel.open(StandardProtocolFamily.INET);
      server_addr = ipv6 ? new InetSocketAddress(getInet6AddressByName(hostname), INITIAL_TID) 
                          : new InetSocketAddress(getInet4AddressByName(hostname), INITIAL_TID);
      data_channel.connect(server_addr);
      data_channel.configureBlocking(false);
      INTEREST_OPS = SelectionKey.OP_WRITE;
      selection_key = data_channel.register(selector, INTEREST_OPS); // initial mode: write
      selection_key.attach(new ConnectionData());
      
      File[] file_list = (new File("input_files/")).listFiles();
      for(int i = 0; i < file_list.length; i++) {
        System.out.println(file_list[i].getName());
        for(int j = 0; j < 5; j++) { // multiple attempts
          loadFile(file_list[i]);
          System.out.println("SENDING FILE " + i + ": TRIAL " + j + ".  DATA SIZE = " + data.size());
          send();
          mode = MODE_WAIT;
        }
      }
    } catch (PortUnreachableException ex) {
      endtime = System.nanoTime();
      //System.out.println("Exiting..."); // server exited, but ack never recieved due to simulated drop
    } catch (Exception ex) {
      System.out.println("  Something went wrong, captain!");
      ex.printStackTrace();
    }
    
    float millis = totaltime / 1000000;
    int bits = totallen * 8;
    float numerator = bits * 1000;
    float denominator = millis * 1000000;
    System.out.println("FINAL: Time taken (ms): " + millis);
    System.out.println("FINAL: Throughput (Mbps): " + (numerator / denominator)); 
  }
  
  public void handleKey(SelectionKey sk) throws Exception {
    //System.out.println("  Event found.  Mode " + mode);
    if(sk.isReadable()) {
      // read
      processAck(sk);
      //sk.interestOps(SelectionKey.OP_WRITE);     
    }
    
    else if(sk.isWritable()) {
      // ack
      sendRequest(sk);
      //sk.interestOps(SelectionKey.OP_READ); 
    }
    
    else return; // bad key
  }

  public void sendRequest(SelectionKey sk) throws Exception {
    //System.out.println("    Sending request...");
    DatagramChannel data_c = (DatagramChannel)sk.channel();
    ConnectionData cd = (ConnectionData)sk.attachment();
    cd.address = server_addr; // pull in data and set reply address
    ByteBuffer info = cd.request;
    cd.request.rewind();
    
    if(mode == MODE_WAIT) { // wait for a request
      // send a request
      //mode = MODE_WRITE;
      // send filename
      //filename = "the_target_file.png";
      byte[] target_file = filename.getBytes("ISO-8859-1");
      int bytelength = WRQ.length + target_file.length + 1 + OCTET.length + 1;
      //System.out.println(bytelength);
      byte[] info2 = new byte[bytelength];
      cd.request = ByteBuffer.allocate(bytelength);
      info = cd.request; 
      info2[0] = WRQ[0];
      info2[1] = WRQ[1];
      int i = 2;
      for(int j = 0; j < target_file.length; j++) {
        info2[i] = target_file[j];
        i++;
      }
      info2[i] = ZERO_BYTE;
      i++;
      // octet
      for(int j = 0; j < OCTET.length; j++) {
        info2[i] = OCTET[j];
        i++;
      }
      info2[i] = ZERO_BYTE;
      //info = ByteBuffer.allocate(info2.length);
      info.rewind();
      info.put(info2);
      //for(i = 0; i < info2.length; i++) System.out.println(info.get(i));
      //cd.request = info;
      //System.out.println("    Sending WRQ for " + filename);
      
      // set up 
      //for(i = 0; i < info2.length; i++) System.out.println(cd.request.get(i));
      cd.request.rewind();
      int sz = data_c.write(cd.request);//, cd.address);
      //System.out.println("Sent " + sz + " bytes.");
      
      
    }
    
       
    else if (mode == MODE_WRITE) { // send data blocks
      //cd.request = ByteBuffer.allocate(BLOCK_SIZE + 4);
      //System.out.println(blockNum);
      if(blockNum > (LAR + SWS) || blockNum >= data.size() + 1) blockNum = LAR;
      //System.out.println(blockNum);
      int size = data.get(blockNum-1).length;
      cd.request = ByteBuffer.allocate(size + 7);
      
      cd.request.put(DATA);
      
      cd.request.put((byte)(blockNum >>> 8));
      cd.request.put((byte)blockNum);
      
      if(data.get(blockNum-1).length < BLOCK_SIZE) cd.request.put(ONE_BYTE); // sentinel byte
      else cd.request.put(ZERO_BYTE);
      
      cd.request.put((byte)(size >>> 8));
      cd.request.put((byte)size);
      
      cd.request.put(data.get(blockNum-1));
      
      cd.request.rewind();
      int sz = data_c.write(cd.request);
      //System.out.println("    Block " + blockNum + ": " + sz + " bytes sent.");
      blockNum++;
      if(LAR == data.size() && sentframe[data.size()-1] == true) mode = MODE_DONE;
    }
    
    
    // do something with data in cd.request
    // set cd.response
  }
  
  public void processAck(SelectionKey sk) throws Exception {
    //System.out.println("    Processing ack...");
    DatagramChannel data_c = (DatagramChannel)sk.channel();
    ConnectionData cd = (ConnectionData)sk.attachment();
    cd.response.rewind();
    data_c.receive(cd.response);
    if(fakedrop && random.nextDouble() <= DROPRATE) return;
    
    int seqNum = (cd.response.get(2) << 8) | (cd.response.get(3));
    if(seqNum == 0 && mode == MODE_WAIT) {
      mode = MODE_WRITE;
    } else {
      //System.out.println("    Recieved ack for seqNum " + seqNum + " (LAR = " + LAR + ")");

      if(seqNum >= LAR  && seqNum <= (LAR + SWS)) {
        markSent(seqNum);
        LAR = determineLAR();
        maintainSenderInvariant();
      }
    }
  }
  
  public void markSent(int framenum) {
    if(framenum < 1) return;
    int i = 0;
    while(i < framenum) {
      sentframe[i] = true;
      i++;
    }
  }
  
  public int determineLAR() {
    int i;
    for(i = 0; i < sentframe.length; i++) {
      //System.out.println("i = " + i);
      if(sentframe[i] == false) return i+1;
    }
    return i;
  }
}
