// Simple Java Easy Command Line Argument Parser - JEclap
// Written by Jacob Peck
// version 20110515

import java.util.Scanner;
import java.util.ArrayList;


/** JEclap: the Java Easy Command Line Argument Parser
 *
 *  @author Jacob Peck (<a href="http://suspended-chord.info">http://suspended-chord.info</a>)
 *  @version 20110515
 */
public class JEclap {
  
  // member variables
  private ArrayList<String> args = new ArrayList<String>();
  
  
  /** JEclap constructor.
   *
   *  @param argv    A String array to operate on.
   */
  public JEclap (String[] argv) {
    // initialize ArrayList
    for (String s : argv) {
      args.add(s);
    }
  }
  
  // methods
  
  /** Checks for the presense of an argument in <code>argv</code>.
   *
   *  @param arg        The argument to look for, minus the preceeding <code>-</code>.
   *  @return           Boolean, true if <code>-arg</code> is in <code>argv</code>; else, returns false.
   */
  public boolean hasArg(String arg) {
    return getFlag(arg);
  }
  
  /** Retrieve all arguments.
   *
   *  @return           A String containing all arguments in <code>argv</code>.
   */
  public String getAllArgs() {
    String temp = "";
    for (String s : args) {
      temp = temp + " " + s;
    }
    return temp.trim();
  }
  
  /** Poll for a singleton argument value.
   *
   *  @param arg        The argument to look for, minus the preceeding <code>-</code>.
   *  @return           The String value associated with <code>-arg</code>, else <code>null</code>.
   */
  public String getArg(String arg) {
    arg = "-" + arg;
    int index = args.indexOf(arg);
    if (index < 0 || args.size() < index+2)
      return null;
    else
      return args.get(index+1);
  }
  
  /** Same as <code>hasArg()</code>. */
  public boolean getFlag(String arg) {
    arg = "-" + arg;
    return args.contains(arg);
  }
  
  /** Poll for a double argument value. (really just a call to <code>getArgNTuple(arg, 2)</code>)
   *
   *  @param arg        The argument to look for, minus the preceeding <code>-</code>.
   *  @return           A String array containing the two values associated with <code>-arg</code>, else <code>null</code>.
   */
  public String[] getArgDouble(String arg) {
    return getArgNTuple(arg, 2);
  }
  
/** Poll for an N-tuple argument value.
   *
   *  @param arg        The argument to look for, minus the preceeding <code>-</code>.
   *  @param n          The number of values to look for.
   *  @return           A String array containing the <code>n</code> values associated with <code>-arg</code>, else <code>null</code>.
   */
  public String[] getArgNTuple(String arg, int n) {
    String[] value = new String[n];
    arg = "-" + arg;
    int index = args.indexOf(arg);
    if (index < 0 || args.size() < index + (n + 1))
      return null;
    else
      for (int i = 0; i < n; i++)
        value[i] = args.get(++index);
    return value;
  }
}