// TFTP Sliding Window Server
// Jacob M. Peck
// CSC 445 - Lea

// note: some code adapted from: 
// http://thushw.blogspot.com/2011/06/asynchronous-udp-server-using-java-nio.html

import java.nio.*;
import java.nio.channels.*;
import java.util.*;
import java.io.*;
import java.net.*;

public class Server extends TFTPConstants {
  public int mode = MODE_WAIT;
  public ArrayList<byte[]> data = new ArrayList<byte[]>();
  public String filename = "";
  public boolean lastframe = false;
  public boolean ipv6;
  public boolean fakedrop;
  public final double DROPRATE = 0.01;
  public Random random = new Random();
  public Selector selector;
  public SelectionKey selection_key;
  
  public static void main(String[] args) {
    Server s = new Server();
    JEclap jeclap = new JEclap(args);
    // args
    if(jeclap.hasArg("6")) s.ipv6 = true; // ipv6
    if(jeclap.hasArg("d")) s.fakedrop = true; // fake drop
    if(jeclap.hasArg("s")) s.RWS = 8; // sliding window
    s.run();
  }
  
  public void read() throws Exception {
    initializeReciever();
    while(mode != MODE_DONE) {
      //System.out.println("Waiting on events.  Mode " + mode);
      if(mode != MODE_WAIT) selector.select(TIMEOUT);
      else selector.select();
      //System.out.println("  Selected.");
      for(SelectionKey sk : selector.selectedKeys()) {
        //System.out.println("  Event!");
        selector.selectedKeys().remove(sk);
        
        if(!sk.isValid()) continue; // bad key
        
        handleKey(sk); // handle the key
      }
      if(mode != MODE_WAIT) toggleInterestOps();
      selection_key.interestOps(INTEREST_OPS);
    }
    
    // write file
    File outfile = new File("output_files" + File.separatorChar + filename);
    FileOutputStream fout = new FileOutputStream(outfile);
    for(byte[] b : data) fout.write(b);
    fout.close();
    
    mode = MODE_WAIT;
    lastframe = false;
    data = new ArrayList<byte[]>();
    INTEREST_OPS = SelectionKey.OP_READ;
    //System.out.println("Exiting...");
  }
  
  public void run() {
    try {
      System.out.println("Initializing...");
      initializeReciever();
      maintainRecieverInvariant();
      
      
      selector = Selector.open();
      DatagramChannel data_channel = null;
      data_channel = ipv6 ? DatagramChannel.open(StandardProtocolFamily.INET6)
                          : DatagramChannel.open(StandardProtocolFamily.INET);
      InetSocketAddress bindaddr = ipv6 ? new InetSocketAddress(getInet6AddressByName("any"), INITIAL_TID) 
                                        : new InetSocketAddress(getInet4AddressByName("any"), INITIAL_TID);
      System.out.println("bindaddr: " + bindaddr);
      data_channel.socket().bind(bindaddr); // port 2700
      data_channel.configureBlocking(false);
      INTEREST_OPS = SelectionKey.OP_READ;
      selection_key = data_channel.register(selector, INTEREST_OPS); // initial mode: read
      selection_key.attach(new ConnectionData());
      
      for(;;) read();

    } catch (Exception ex) {
      System.out.println("  Something went wrong, captain!");
      ex.printStackTrace();
    }
  }
  
  public void handleKey(SelectionKey sk) throws Exception {
    //System.out.println("  Event found.  Mode " + mode);
    if(sk.isReadable()) {
      // read
      processRequest(sk);
      sk.interestOps(SelectionKey.OP_WRITE);     
    }
    
    else if(sk.isWritable()) {
      // ack
      sendAck(sk);
      sk.interestOps(SelectionKey.OP_READ); 
    }
    
    else return; // bad key
  }

  public void processRequest(SelectionKey sk) throws Exception {
    //System.out.println("    Processing request...");
    DatagramChannel data_c = (DatagramChannel)sk.channel();
    ConnectionData cd = (ConnectionData)sk.attachment();
    //cd.request.rewind();
    cd.request = ByteBuffer.allocate(BLOCK_SIZE + 7);
    
    cd.address = data_c.receive(cd.request); // pull in data and set reply address
    //System.out.println(cd.address);
    //data_c.connect(cd.address);
    ByteBuffer info = cd.request;
    //byte[] info_array = new byte[1024];
    //info.get(info_array);
    
    //for(int i = 0; i < 4; i++) System.out.println(info.get(i));
        
    if(mode == MODE_WAIT) { // wait for a request
      //System.out.println("    Mode = MODE_WAIT");
      //System.out.println(info.get(0) + info.get(1)); 
      if(info.get(0) == WRQ[0] && info.get(1) == WRQ[1]) { // wrq
        mode = MODE_READ;
        // read filename
        List<Byte> fname_bytes = new ArrayList<Byte>();
        int i = 2;
        byte b = info.get(i);
        while(b != ZERO_BYTE) {
          fname_bytes.add(b);
          i++;
          b = info.get(i);
        }
        byte[] fname_bytes2 = new byte[fname_bytes.size()];
        for(i = 0; i < fname_bytes2.length; i++) {
          fname_bytes2[i] = fname_bytes.get(i);
        }
        try {
          filename = new String(fname_bytes2, "ISO-8859-1");
        } catch (Exception ex) {
          ex.printStackTrace();
          System.exit(-1);
        }
        // assume octet
        //System.out.println("    Received WRQ for " + filename);
        
        // set up ack
        cd.response = ByteBuffer.allocate(4);
        cd.response.put(ACK);
        cd.response.put((byte)0);
        cd.response.put((byte)0);
      } // else error
    }
    
    else if (mode == MODE_READ) { // read data blocks
      if(fakedrop && random.nextDouble() <= DROPRATE) return;
      //System.out.println("    Mode = MODE_READ: LFR = " + LFR + " LAF = " + LAF + " RWS = " + RWS);     
      //for(int i = 0; i < 516; i++) System.out.println(i + ": " + info.get(i));
      if(info.get(0) == DATA[0] && info.get(1) == DATA[1]) {
        // blocknum
        int seqNum = (info.get(2) << 8) | info.get(3);
        //System.out.println("    seqNum: " + seqNum);
        if(seqNum <= LAF && seqNum > LFR) { // valid frame
          // is final?
          if(info.get(4) == ONE_BYTE) lastframe = true;
          //System.out.println("    Last frame? " + info.get(4));
          // extract size
          int size = ((info.get(5) & 0xFF) << 8) | (info.get(6) & 0xFF);
          //System.out.println(info.get(5) << 8);
          //System.out.println(info.get(6) << 8);
          //System.out.println("    Size: " + size);
          byte[] framecontents = new byte[size];
          info.rewind();
          //System.out.println("    Info about info: pos = " + info.position() + " limit = " + info.limit() + " capacity = " + info.capacity());
          info.position(7);
          info.get(framecontents, 0, framecontents.length);
          //System.out.println("    Info about info: pos = " + info.position() + " limit = " + info.limit() + " capacity = " + info.capacity());
          while(data.size() <= seqNum - 1) data.add(null);
          //System.out.println("    Info about data: size = " + data.size());
          data.set(seqNum - 1, framecontents);
        }
        //System.out.println("    Recieved DATA block " + seqNum);
        
        //for(int i = 0; i < 4; i++) System.out.println(data.get(seqNum)[i]);
        
        // set up ack
        cd.response = ByteBuffer.allocate(4);
        cd.response.put(ACK);
        LFR = seqNumToAck();
        maintainRecieverInvariant();
        cd.response.put((byte)(LFR >> 8));
        cd.response.put((byte)LFR);
      }
      // else error
    }
  }
  
  public int seqNumToAck() {
    int i = 0;
    for(byte[] b : data) {
      //System.out.println(b);
      if(b == null) break;
      i++;
    }
    //i--;
    //System.out.println("    seqNumToAck() = " + i);
    //System.out.println("    data size: " + data.size());
    //System.out.println("    lastframe: " + lastframe);
    
    if(lastframe && i == data.size()) { 
      mode = MODE_DONE; // exit after ack if all has been recieved
      //System.out.println("DONE");
    }
    
    return i;
  }
  
  public void sendAck(SelectionKey sk) throws Exception {
    DatagramChannel data_c = (DatagramChannel)sk.channel();
    ConnectionData cd = (ConnectionData)sk.attachment();
    cd.response.rewind();
    if(cd.address != null) data_c.send(cd.response, cd.address);
    //System.out.println("    Sending ack..." + seqNumToAck());
  }
}
