// TFTP Constants
// Jacob M. Peck
// CSC 445 - Lea

import java.nio.channels.*;
import java.net.*;

public class TFTPConstants {
  // packet format
  public static final byte[] RRQ = {0,1};
  public static final byte[] WRQ = {0,2};
  public static final byte[] DATA = {0,3};
  public static final byte[] ACK = {0,4};
  public static final byte[] ERROR = {0,5};
  
  // request modes
  public static byte[] NETASCII = null;
  public static byte[] OCTET = null;
  public static byte[] MAIL = null;
  static {
    try {
      NETASCII = "netascii".getBytes("ISO-8859-1");
      OCTET = "octet".getBytes("ISO-8859-1");
      MAIL = "mail".getBytes("ISO-8859-1");
    } catch (Exception ex) {
      ex.printStackTrace();
      System.exit(-1);
    }
  }
  
  // block size
  public static final int BLOCK_SIZE = 512; // 512 byte blocksize
  
  // zero byte
  public static final byte ZERO_BYTE = 0;
  public static final byte ONE_BYTE = 1;
  
  // error codes
  public static final byte ERROR_UNDEFINED = 0;
  public static final byte ERROR_FILE_NOT_FOUND = 1;
  public static final byte ERROR_ACCESS_VIOLATION = 2;
  public static final byte ERROR_DISK_FULL = 3;
  public static final byte ERROR_ILLEGAL_OPERATION = 4;
  public static final byte ERROR_UNKNOWN_TID = 5;
  public static final byte ERROR_FILE_EXISTS = 6;
  public static final byte ERROR_USER_NOT_FOUND = 7;
  
  // initial TID
  public static final int INITIAL_TID = 2700; // default 69
  
  // modes for state tracking
  public static final int MODE_WAIT = 0;
  public static final int MODE_READ = 1;
  public static final int MODE_WRITE = 2;
  public static final int MODE_DONE = 3;
  
  // state
  // sender
  public int SWS = 1; // sliding window size
  public int blockNum = 1;
  public int LFS = 0; // largest frame sent
  public int LAR = 0; // last acked frame
  
  public void initializeSender() { blockNum = 1; LFS = 0; LAR = 0; }
  public void maintainSenderInvariant() { LFS = (LAR + SWS); }
  
  // receiver
  public int RWS = 1; // recieving window size
  public int LAF = 1; // largest acceptible frame
  public int LFR = 0; // last frame recieved
  
  public void initializeReciever() { LAF = 1; LFR = 0; }
  public void maintainRecieverInvariant() { LAF = LFR + RWS; }
  
  
  // timeout
  public static final int TIMEOUT = 1;
  
  // current interest ops
  public static int INTEREST_OPS;
  
  public void toggleInterestOps() {
    if(INTEREST_OPS == SelectionKey.OP_WRITE) INTEREST_OPS = SelectionKey.OP_READ;
    else INTEREST_OPS = SelectionKey.OP_WRITE;
  }
  
  // IPvX
  // (from http://stackoverflow.com/a/3872194)
  static Inet6Address getInet6AddressByName(String host) throws UnknownHostException, SecurityException {
    // workarounds
    switch(host) {
      case "any" : host = "::"; break;
      case "localhost" : host = "::1"; break;
      case "moxie" : host = "fe80::214:4fff:fee8:adb4"; break;
      case "lambda" : host = "fe80::214:4fff:fee7:6b22"; break;
      case "altair" : host = "fe80::209:3dff:fe14:64ff"; break;
      case "gee" : host = "fe80::214:4fff:fe98:3244"; break;
    }
    for(InetAddress addr : InetAddress.getAllByName(host))
      if(addr instanceof Inet6Address) return (Inet6Address)addr;
      //System.out.println(addr);
    throw new UnknownHostException("No IPv6 address found for " + host);
  }
  
  static Inet4Address getInet4AddressByName(String host) throws UnknownHostException, SecurityException {
    switch(host) {
      case "any" : host = "0.0.0.0"; break;
    }
    for(InetAddress addr : InetAddress.getAllByName(host))
      if(addr instanceof Inet4Address) return (Inet4Address)addr;
    throw new UnknownHostException("No IPv4 address found for " + host);
  }
}